from typing import Sequence

import numpy as np

from .cpphelpers import lib
from .types import NumberTypes

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


class TatamiNumericPointer:
    """Initialize a Tatami Numeric Ponter object."""

    def __init__(self, ptr: "lib.Mattress"):
        """Initialize the class.

        Args:
            ptr (lib.Mattress): pointer to the tatami object.
        """
        self.ptr = ptr

    def __del__(self):
        lib.py_free_mat(self.ptr)

    def nrow(self) -> int:
        """Get number of rows.

        Returns:
            int: number of rows.
        """
        return lib.py_extract_nrow(self.ptr)

    def ncol(self) -> int:
        """Get number of columns.

        Returns:
            int: number of columns.
        """
        return lib.py_extract_ncol(self.ptr)

    def sparse(self) -> bool:
        """Is the matrix sparse?

        Returns:
            bool: True if matrix is sparse.
        """
        return lib.py_extract_sparse(self.ptr) > 0

    def row(self, r: int) -> Sequence[NumberTypes]:
        """Access a row from the tatami matrix.

        Args:
            r (int): row to access.

        Returns:
            Sequence[NumberTypes]: row from the matrix.
        """
        output = np.ndarray((self.ncol(),), dtype="float64")
        lib.py_extract_row(self.ptr, r, output.ctypes.data)
        return output

    def column(self, c: int) -> Sequence[NumberTypes]:
        """Access a column from the tatami matrix.

        Args:
            c (int): column to access.

        Returns:
            Sequence[NumberTypes]: column from the matrix.
        """
        output = np.ndarray((self.nrow(),), dtype="float64")
        lib.py_extract_column(self.ptr, c, output.ctypes.data)
        return output

    @classmethod
    def from_dense_matrix(
        cls, x: np.ndarray, dtype: str, order: bool
    ) -> "TatamiNumericPointer":
        """Initialize class from a dense matrix.

        Args:
            x (np.ndarray): input numpy matrix.
            dtype (str): dtype of the values.
            order (bool): True if order is 'C' else False.

        Returns:
            TatamiNumericPointer: instance of the class.
        """
        return cls(
            lib.py_initialize_dense_matrix(
                x.shape[0], x.shape[1], dtype, x.ctypes.data, order
            )
        )



__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def map_axis_to_bool(axis: int) -> bool:
    """Validate and map axis to a boolean.

    Args:
        axis (int): 0 for rows or 1 for columns.

    Raises:
        ValueError: If axis is not 0 or 1.

    Returns:
        bool: True if axis is 0 else False.
    """
    if not (axis == 0 or axis == 1):
        raise ValueError(f"Axis must be 0 or 1, provided {axis}")

    return True if axis == 0 else False


def map_order_to_bool(order: str) -> bool:
    """Validate and map order to a boolean.

    Args:
        order (str): dense matrix representation, ‘C’, ‘F’,
            row-major (C-style) or column-major (Fortran-style) order.

    Raises:
        ValueError: If order is not 'C' or 'F'.

    Returns:
        bool: True if order is 'C' else False.
    """
    if not (order == "C" or order == "F"):
        raise ValueError(f"Order must be 'C' or 'F', provided {order}")

    return True if order == "C" else False



import ctypes as ct
import os
from typing import Sequence

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


# TODO: surely there's a better way than whatever this is.
def includes() -> Sequence[str]:
    """Provides access to C++ headers (including tatami) for downstream packages.

    Returns:
        Sequence[str]: list of paths to the header files.
    """
    dirname = os.path.dirname(os.path.abspath(__file__))
    return [
        os.path.join(dirname, "extern", "tatami", "include"),
        os.path.join(dirname, "include"),
    ]


def load_dll() -> ct.CDLL:
    """load the shared library.

    usually starts with core.<platform>.<so or dll>.

    Returns:
        (ct.CDLL, optional): shared object.
    """
    dirname = os.path.dirname(os.path.abspath(__file__))
    contents = os.listdir(dirname)
    for x in contents:
        if x.startswith("core") and not x.endswith("py"):
            return ct.CDLL(os.path.join(dirname, x))

    raise Exception(
        "Cannot find the shared object file! Report this issue to dev on github."
    )


lib = load_dll()

lib.py_free_mat.argtypes = [ct.c_void_p]
lib.py_extract_nrow.restype = ct.c_int
lib.py_extract_nrow.argtypes = [ct.c_void_p]
lib.py_extract_ncol.restype = ct.c_int
lib.py_extract_ncol.argtypes = [ct.c_void_p]
lib.py_extract_sparse.restype = ct.c_int
lib.py_extract_sparse.argtypes = [ct.c_void_p]
lib.py_extract_row.argtypes = [ct.c_void_p, ct.c_int, ct.c_void_p]
lib.py_extract_column.argtypes = [ct.c_void_p, ct.c_int, ct.c_void_p]


lib.py_initialize_dense_matrix.restype = ct.c_void_p
lib.py_initialize_dense_matrix.argtypes = [
    ct.c_int,
    ct.c_int,
    ct.c_char_p,
    ct.c_void_p,
    ct.c_char,
]



from functools import singledispatch
from typing import Any

import numpy as np

from .TatamiNumericPointer import TatamiNumericPointer
from .utils import map_order_to_bool

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


@singledispatch
def tatamize(x: Any, order: str = "C") -> TatamiNumericPointer:
    """Converts python matrix representations to tatami.

    Args:
        x (Any): Any matrix-like object.
        order (str): dense matrix representation, ‘C’, ‘F’,
            row-major (C-style) or column-major (Fortran-style) order.

    Raises:
        NotImplementedError: if x is not supported.

    Returns:
        TatamiNumericPointer: a pointer to tatami object.
    """
    raise NotImplementedError(
        f"tatamize is not supported for objects of class: {type(x)}"
    )


@tatamize.register
def _tatamize_numpy(x: np.ndarray, order: str = "C") -> TatamiNumericPointer:
    """Converts numpy representations to tatami.

    Args:
        x (np.ndarray): A numpy nd-array object.
        order (str): dense matrix representation, ‘C’, ‘F’,
            row-major (C-style) or column-major (Fortran-style) order.

    Raises:
        NotImplementedError: if x is not supported.

    Returns:
        TatamiNumericPointer: a pointer to tatami object.
    """
    order_to_bool = map_order_to_bool(order=order)
    dtype = str(x.dtype).encode("utf-8")
    return TatamiNumericPointer.from_dense_matrix(x, dtype=dtype, order=order_to_bool)



import numpy as np
from mattress import tatamize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_numpy():
    y = np.random.rand(1000, 100)
    ptr = tatamize(y)
    assert all(ptr.row(0) == y[0, :])
    assert all(ptr.column(1) == y[:, 1])


def test_numpy_with_order():
    y = np.ones((2, 3), order='F')
    ptr = tatamize(y, order="F")
    assert all(ptr.row(0) == y[0, :])
    assert all(ptr.column(1) == y[:, 1])


def test_numpy_with_dtype():
    y = (np.random.rand(50, 12) * 100).astype("i8")
    ptr = tatamize(y)
    assert all(ptr.row(0) == y[0, :])
    assert all(ptr.column(1) == y[:, 1])
